/*------------------------------------------------------------------------------
	Function to restore object addresses to offsets within a package.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "ObjHeader.h"
#include "PackageParts.h"


//	Determine whether a Newton ref is a pointer (could be an immediate)

inline Boolean IsPtr(Ref ref)
{
	return ISREALPTR(ref) && ((ref & 0xFF000000) != 0x00);
}


//	Fix up the pointer ref offset from base by delta

static void FixRef(Ptr base, Ref ref, long delta)
{
//	point to the object - could  be frame or array as well as binary
	ArrayObject * rObj = (ArrayObject *) (base + (ULong) RPTR(ref));
	Ref fixedRef;

//	point to its class
	Ref * refPointer = &(rObj->objClass);
	if (IsPtr(*refPointer))
	{
		fixedRef = (Ref) (RPTR(*refPointer) - delta);
		*refPointer = MAKEPTR(fixedRef);					//	update the class ref
		if (rObj->flags & kObjMask == kFrameObject)
			FixRef(base, fixedRef, delta);				//	for frames, class is actually map
	}

//	if its a frame / array, step through each slot / element fixing those
	if ((rObj->flags & kObjSlotted))
	{
		refPointer = &rObj->slot[0];
		for (FastInt refCounter = ARRAYLENGTH(rObj);  refCounter > 0; refCounter--, refPointer++)
		{
			if (IsPtr(*refPointer))
			{
				fixedRef = (Ref) (RPTR(*refPointer) - delta);
				if (fixedRef > 0)
				{
					*refPointer = MAKEPTR(fixedRef);		//	update this ref
					FixRef(base, fixedRef, delta);		//	follow it to update sub-refs
				}
			}
		}
	}
}


/*------------------------------------------------------------------------------
	Reset the relocation offsets in a package to be relative to the package
	and not absolute addresses.
	We do this by following object refs from each part base and fixing up
	the addresses.

	Args:		pkg
					package binary object

	Return:	the modified package
------------------------------------------------------------------------------*/

extern "C"
Ref UnRelocatePkg(RefArg rcvr, RefArg pkg)
{
	//	parameter must be a binary object
	if (!IsBinary(pkg))
		ThrowBadTypeWithFrameData(kNSErrNotAPackage, pkg);

	WITH_LOCKED_BINARY(pkg, p)

	Ptr pp = (Ptr) p;
	PackageDirectory * dir = (PackageDirectory *) pp;
	ULong partData = dir->directorySize;

//	if its a "package1" with relocation info then adjust that relocation info
	if ((dir->signature[7] == '1') && (dir->flags & kRelocationFlag))
	{
		RelocationHeader * hdr = (RelocationHeader *) (pp + partData);
		long delta = (long) pp - hdr->baseAddress;
		RelocationSet * set = (RelocationSet *) (hdr + 1);
		for (FastInt setNumber = 0; setNumber < hdr->numEntries; setNumber++)
		{
			long * pageBase = (long *) (pp + set->pageNumber * hdr->pageSize);
			for (FastInt offsetNumber = 0; offsetNumber < set->offsetCount; offsetNumber++)
				*(pageBase + set->offsets[offsetNumber]) -= delta;
			set = (RelocationSet *) ((char *) set
					+ sizeof(UShort) * 2					// RelocationSet less offsets
					+ LONGALIGN(set->offsetCount));
		}
		partData += hdr->relocationSize;
	}

//	calculate delta to be applied to refs
	ArrayObject * rObj = (ArrayObject *) (pp + partData);
	long delta = (long) RPTR(rObj->slot[0])		//	address of first object
					- (partData + rObj->size);			//	offset of first object

//	fix up refs for all parts
	for (FastInt partNumber = 0; partNumber < dir->numParts; partNumber++)
		FixRef(pp, partData + dir->parts[partNumber].offset, delta);

	END_WITH_LOCKED_BINARY(pkg)

	return pkg;
}
